classdef VLMC < Context.Kernel
    % piecewise constant kernel
    % determined by a complete trie and the distributions at its leaves
    %
    % author: Aurlien Garivier
    % reference: Perfect Simulation Scheme For Processes With Long Memory :
    %    A ``Coupling Into And From The Past'' Algorithm, by A. Garivier    
    
    properties
        CSD % Complete Suffix Trie
        distribution % for a leaf, the conditional distribution; otherwise, the sub-distribution common to all leaves below the node
        partition % defines the coupling function, see coefficients alpha and beta in the article
    end

    
    methods
        function obj = VLMC(contexts, distribs)
            % given a cell array of contexts and a cell array of
            % distributions (vectors), builds the corresponding VLMC
            obj.K = max([contexts{:}]);
            obj.CSD = Context.Trie(obj.K, contexts);
            obj.distribution = cell(1, length(obj.CSD.father));
            leaves = obj.CSD.getLeaves();
            [obj.distribution{leaves}] = distribs{:};
            obj.checkDistributions();
            obj.checkPartition();
        end
        
        function d = checkDistributions(obj, node)
            % updates obj.distribution according to Doeblin's rule
            % and returns the corresponding sub-distribution
            if nargin<2, node = obj.CSD.getRoot(); end 
            if obj.CSD.isLeaf(node)
                d = obj.distribution{node};
            else
                p = zeros(obj.CSD.K); 
                for x = 1:obj.CSD.K
                    child = obj.CSD.child(x, node);
                    p(x, :) = obj.checkDistributions(child);
                end
                d = min(p, [], 1);
                obj.distribution{node} = d;
            end
        end
        
        function checkPartition(obj)
            % updates obj.partition to build to the coupling of the article
            node = obj.CSD.getRoot();
            obj.partition{node} = cumsum([0, obj.distribution{node}]);
            if ~obj.CSD.isLeaf(node), toCheck = obj.CSD.child(:, node); else toCheck = []; end
            while ~isempty(toCheck)
                node = toCheck(1); toCheck = toCheck(2:end);
                father = obj.CSD.getFather(node);
                obj.partition{node} = obj.partition{father}(end) + [0, cumsum(obj.distribution{node}-obj.distribution{father})];
                if ~obj.CSD.isLeaf(node)
                    toCheck = [toCheck; obj.CSD.child(:, node)];
                end
            end
        end
                
        function f = transition(obj, u, node)
            % given u in [0,1[, returns the PCfunction of the context 
            % giving the next symbol
            % best implementation: builds directly the minimal trie
            if nargin<3, node = obj.CSD.getRoot(); end
            f = Context.PCfunction(Context.Trie(obj.CSD.K), {[]});
            if u<obj.partition{node}(end)
                f.values{1} = sum(u>obj.partition{node});
            else
                for x=1:obj.CSD.K
                    ftemp = obj.transition(u, obj.CSD.child(x, node));
                    f.CSD.addSubTrie(f.CSD.getRoot(), x, ftemp.CSD);
                    f.values = {f.values{:}, ftemp.values{:}};
                end
            end
        end

        function f = transition2(obj, u)
            % given u in [0,1[, returns the PCfunction of the context 
            % giving the next symbol
            % most simple (and less efficient) implementation: pruning of
            % obj.CSD
            v = obj.next(u);
            f = Context.PCfunction(obj.CSD.copy(), v);
            f.prune();
        end
        
        function v = next(obj, u, nodes)
            % given u in [0,1[, returns {phi(u, node)} for all node in the 
            % list 'nodes'
            if nargin<3, nodes = 1:length(obj.CSD.father); end
            v = cell(1, length(nodes));
            for k=1:length(nodes)
               if u>=obj.partition{nodes(k)}(1)
                   v{k} = sum(u>obj.partition{nodes(k)});
               else
                   v{k} = cell2mat(obj.next(u, obj.CSD.getFather(nodes(k))));
               end
            end
        end
        
        function newIndex = reorder(obj)
            % re-orders children so that, in the plot, the order of children
            % from left to right is the same as the order of labels
            % modifies obj.values  accordingly
            % returns perm: after reorder, node k is now perm(k)
            newIndex = obj.CSD.reorder();
            [obj.distribution{newIndex}] = obj.distribution{:};
            [obj.partition{newIndex}] = obj.partition{:};
        end
        
        function [x,y]=plot(obj, displayPartition)
            if nargin<2, displayPartition = false; end
            obj.reorder();
            [x,y] = obj.CSD.plot();
            for j=1:length(obj.CSD.father)
                if displayPartition
                    text(x(j)+0.01, y(j), Context.part2str(obj.partition{j})); 
                else
                    text(x(j)+0.01, y(j), Context.dist2str(obj.distribution{j})); 
                end
            end
        end
    end
end

